#include "handlers.h"

#include <dlfcn.h>
#include <ffi.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include "net.h"
#include "server.h"


#define CP_TYPE_SCHAR   0
#define CP_TYPE_UCHAR   1
#define CP_TYPE_SSHORT  2
#define CP_TYPE_USHORT  3
#define CP_TYPE_SINT    4
#define CP_TYPE_UINT    5
#define CP_TYPE_SLONG   6
#define CP_TYPE_ULONG   7
#define CP_TYPE_STRUCT  27

#define CP_ABI_DEFAULT 0

#define ARRAY_LEN(array) (sizeof(array) / sizeof(array[0]))

#define MAX(a, b) ((a) > (b) ? (a) : (b))

union flexer
{
    uint8_t u8;
    uint16_t u16;
    uint32_t u32;
    uint64_t u64;
};

static ffi_type* direct_type_conversion_array_g[] = {
    &ffi_type_schar,
    &ffi_type_uchar,
    &ffi_type_sshort,
    &ffi_type_ushort,
    &ffi_type_sint,
    &ffi_type_uint,
    &ffi_type_slong,
    &ffi_type_ulong,
    &ffi_type_sint8,
    &ffi_type_uint8,
    &ffi_type_sint16,
    &ffi_type_uint16,
    &ffi_type_sint32,
    &ffi_type_uint32,
    &ffi_type_sint64,
    &ffi_type_uint64,
    &ffi_type_pointer,
};

static ffi_type* recv_type(int fd);
static ffi_type* recv_struct(int fd);
static void free_ffi_type(ffi_type* type);
static int interpret_abi(uint8_t passed_abi, ffi_abi* f_abi);

/**
 * Load a shared library and send back the handle
 */
int cp_load(struct cpype* pype)
{
    uint32_t libname_len;
    char* libname = NULL;
    void* lib = NULL;
    int rtn = -1;
    fputs("[+] Entering cp_load\n", stderr);
    if (recv_n(pype->fin, &libname_len, sizeof(libname_len)))
    {
        fputs("[-] Failed to receive libname length\n", stderr);
        goto end;
    }
    libname_len = ntohl(libname_len);
    libname = calloc(libname_len + 1, 1);
    if (!libname)
    {
        // Failed to alloc. Still send NULL response.
        fputs("[-] Failed to alloc libname\n", stderr);
        goto send_lib;
    }
    if (recv_n(pype->fin, libname, libname_len))
    {
        fputs("[-] Failed to receive libname\n", stderr);
        goto end;
    }
    lib = dlopen(libname, RTLD_NOW);
    fprintf(stderr, "[+] Found lib '%s' at %p\n", libname, lib);
    lib = HTONP(lib);
send_lib:
    if (send_n(pype->fout, &lib, sizeof(lib)))
    {
        fputs("[-] Failed to send lib pointer\n", stderr);
        goto end;
    }
    rtn = 0;
end:
    free(libname);
    return rtn;
}

int cp_find(struct cpype* pype)
{
    void* lib;
    uint32_t symname_len;
    char* symname = NULL;
    void* sym = NULL;
    int rtn = -1;
    fputs("[+] Entering cp_find\n", stderr);
    if (recv_n(pype->fin, &lib, sizeof(lib)))
    {
        fputs("[-] Failed to recieve lib pointer\n", stderr);
        goto end;
    }
    lib = NTOHP(lib);
    if (recv_n(pype->fin, &symname_len, sizeof(symname_len)))
    {
        fputs("[-] Failed to receive symname length\n", stderr);
        goto end;
    }
    symname_len = ntohl(symname_len);
    symname = calloc(symname_len + 1, 1);
    if (!symname)
    {
        // Failed to alloc. Still send NULL response.
        fputs("[-] Failed to alloc symname\n", stderr);
        goto send_sym;
    }
    if (recv_n(pype->fin, symname, symname_len))
    {
        fputs("[-] Failed to receive symname\n", stderr);
        goto end;
    }
    sym = dlsym(lib, symname);
    sym = HTONP(sym);
send_sym:
    if (send_n(pype->fout, &sym, sizeof(sym)))
    {
        fputs("[-] Failed to send symbol pointer\n", stderr);
        goto end;
    }
    rtn = 0;
end:
    free(symname);
    return rtn;
}

int cp_read(struct cpype* pype)
{
    int rtn = -1;
    void* address;
    union flexer flex;
    ffi_type* type;
    fputs("[+] Entering cp_read\n", stderr);
    type = recv_type(pype->fin);
    if (!type)
    {
        fputs("[-] Failed to receive type\n", stderr);
        goto end;
    }
    if (type->type == FFI_TYPE_STRUCT)
    {
        fputs("[-] Cannot read directly from a struct\n", stderr);
        goto end;
    }
    if (recv_n(pype->fin, &address, sizeof(address)))
    {
        fputs("[-] Failed to receive address\n", stderr);
        goto end;
    }
    address = NTOHP(address);
    switch (type->size)
    {
        case sizeof(uint8_t):
            flex.u8 = *((uint8_t*) address);
            break;
        case sizeof(uint16_t):
            flex.u16 = htobe16(*((uint16_t*) address));
            break;
        case sizeof(uint32_t):
            flex.u32 = htobe32(*((uint32_t*) address));
            break;
        case sizeof(uint64_t):
            flex.u64 = htobe64(*((uint64_t*) address));
            break;
        default:
            fprintf(stderr, "[-] Cannot handle size %zd\n", type->size);
            goto end;
    }
    if (send_n(pype->fout, &flex, type->size))
    {
        fputs("[-] Failed to send read data\n", stderr);
        goto end;
    }
    rtn = 0;
end:
    return rtn;
}

int cp_write(struct cpype* pype)
{
    int rtn = -1;
    void* address;
    ffi_type* type;
    fputs("[+] Entering cp_write\n", stderr);
    type = recv_type(pype->fin);
    if (!type)
    {
        // TODO: We might want to send back a failure message.
        fputs("[-] Failed to receive type\n", stderr);
        goto end;
    }
    if (type->type == FFI_TYPE_STRUCT)
    {
        fputs("[-] Cannot write directly to a struct\n", stderr);
        goto end;
    }
    if (recv_n(pype->fin, &address, sizeof(address)))
    {
        fputs("[-] Failed to receive address\n", stderr);
        goto end;
    }
    address = NTOHP(address);
    // This is a little complicated because we need to read and
    // potentially byte-swap a variable number of bytes in a portable
    // way. I think the simplest way to do this is to write the bytes
    // to their final destination, then apply a swapping function in
    // place.
    if (recv_n(pype->fin, address, type->size))
    {
        fputs("[-] Failed to receive data to write\n", stderr);
        goto end;
    }
    switch (type->size)
    {
        case sizeof(uint8_t):
            break;
        case sizeof(uint16_t):
            *((uint16_t*) address) = be16toh(*((uint16_t*) address));
            break;
        case sizeof(uint32_t):
            *((uint32_t*) address) = be32toh(*((uint32_t*) address));
            break;
        case sizeof(uint64_t):
            *((uint64_t*) address) = be64toh(*((uint64_t*) address));
            break;
        default:
            fprintf(stderr, "[-] Cannot handle size %zd\n", type->size);
            goto end;
    }
    rtn = 0;
end:
    return rtn;
}

int cp_size(struct cpype* pype)
{
    int rtn = -1;
    size_t size;
    ffi_type* type;
    fputs("[+] Entering cp_size\n", stderr);
    type = recv_type(pype->fin);
    if (!type)
    {
        // TODO: We might want to send back a failure message.
        fputs("[-] Failed to receive type\n", stderr);
        goto end;
    }
    size = HTONZ(type->size);
    if (send_n(pype->fout, &size, sizeof(size)))
    {
        fputs("[-] Failed to send size\n", stderr);
        goto end;
    }
    rtn = 0;
end:
    if (type)
    {
        free_ffi_type(type);
    }
    return rtn;
}

int cp_offsets(struct cpype* pype)
{
    int rtn = -1;
    ffi_status status;
    size_t num_fields;
    size_t* offsets = NULL;
    ffi_type* type;
    fputs("[+] Entering cp_offsets\n", stderr);
    type = recv_type(pype->fin);
    if (!type)
    {
        // TODO: We might want to send back a failure message.
        fputs("[-] Failed to receive type\n", stderr);
        goto end;
    }
    for (num_fields = 0; type->elements[num_fields]; num_fields++);
    offsets = calloc(num_fields, sizeof(size_t));
    if (!offsets)
    {
        fputs("[-] Failed to allocate offsets\n", stderr);
        goto end;
    }
    status = ffi_get_struct_offsets(FFI_DEFAULT_ABI, type, offsets);
    if (status != FFI_OK)
    {
        fputs("[-] Failed to layout struct\n", stderr);
        goto end;
    }
    size_t i;
    for (i = 0; i < num_fields; i++)
    {
        offsets[i] = HTONZ(offsets[i]);
        if (send_n(pype->fout, &offsets[i], sizeof(offsets[i])))
        {
            fprintf(stderr, "[-] Failed to send offset %zd\n", i);
            goto end;
        }
    }
    rtn = 0;
end:
    if (type)
    {
        free_ffi_type(type);
    }
    if (offsets)
    {
        free(offsets);
    }
    return rtn;
}

int cp_alloc(struct cpype* pype)
{
    int rtn = -1;
    size_t size;
    void* ptr;
    void* ptr_nbo;
    fputs("[+] Entering cp_alloc\n", stderr);
    if (recv_n(pype->fin, &size, sizeof(size)))
    {
        fputs("[-] Failed to receive size\n", stderr);
        goto end;
    }
    size = NTOHZ(size);
    ptr = calloc(size, 1);
    if (!ptr)
    {
        fputs("[-] Failed to alloc\n", stderr);
        goto end;
    }
    fprintf(stderr, "[+] Allocated %zd bytes\n", size);
    ptr_nbo = NTOHP(ptr);
    if (send_n(pype->fout, &ptr_nbo, sizeof(ptr_nbo)))
    {
        fputs("[-] Failed to send pointer\n", stderr);
        goto end;
    }
    rtn = 0;
end:
    return rtn;
}

int cp_free(struct cpype* pype)
{
    int rtn = -1;
    void* ptr;
    fputs("[+] Entering cp_free\n", stderr);
    if (recv_n(pype->fin, &ptr, sizeof(ptr)))
    {
        fputs("[-] Failed to receive pointer\n", stderr);
        goto end;
    }
    free(NTOHP(ptr));
    rtn = 0;
end:
    return rtn;
}

int cp_call(struct cpype* pype)
{
    int rtn = -1;
    void* func_ptr;
    uint8_t passed_abi;
    uint32_t num_args;
    uint32_t i;

    void** args;
    ffi_cif cif;
    ffi_type** arg_types;
    ffi_type* rtn_type;
    ffi_abi f_abi;
    ffi_status status;
    void* result;
    size_t result_size;

    fputs("[+] Entering cp_call\n", stderr);

    if (recv_n(pype->fin, &func_ptr, sizeof(func_ptr)))
    {
        fputs("[-] Failed to receive func pointer\n", stderr);
        goto end;
    }
    func_ptr = NTOHP(func_ptr);
    // TODO: Handle different ABI values.
    if (recv_n(pype->fin, &passed_abi, sizeof(passed_abi)))
    {
        fputs("[-] Failed to receive abi\n", stderr);
        goto end;
    }
    if (interpret_abi(passed_abi, &f_abi))
    {
        fprintf(stderr, "[-] Failed to interpret abi: %u\n", passed_abi);
        goto end;
    }

    rtn_type = recv_type(pype->fin);
    if (!rtn_type)
    {
        fputs("[-] Failed to receive rtn_type\n", stderr);
        goto end;
    }

    // Allocate result.
    result_size = MAX(rtn_type->size, sizeof(ffi_arg));

    if (recv_n(pype->fin, &num_args, sizeof(num_args)))
    {
        fputs("[-] Failed to receive num args\n", stderr);
        goto end;
    }
    num_args = be32toh(num_args);
    fprintf(stderr, "[ ] %u args\n", num_args);

    arg_types = calloc(num_args + 1, sizeof(*arg_types));
    args = calloc(num_args, sizeof(*args));
    if (!args || !arg_types)
    {
        fputs("[-] Failed to alloc args or arg_types\n", stderr);
        goto end;
    }

    for (i = 0; i < num_args; i++)
    {
        arg_types[i] = recv_type(pype->fin);
        if (!arg_types[i])
        {
            fprintf(stderr, "[-] Failed to receive arg type %d\n", i);
            goto end;
        }
    }
    for (i = 0; i < num_args; i++)
    {
        if (recv_n(pype->fin, args + i, sizeof(*args)))
        {
            fprintf(stderr, "[-] Failed to receive arg %d\n", i);
            goto end;
        }
        args[i] = NTOHP(args[i]);
    }

    status = ffi_prep_cif(&cif, f_abi,
        num_args, rtn_type, arg_types);
    if (status != FFI_OK)
    {
        fputs("[-] ffi_prep_cif failed\n", stderr);
        goto end;
    }

    result = malloc(result_size);
    if (!result)
    {
        fputs("[-] Failed to alloc result\n", stderr);
        goto end;
    }

    ffi_call(&cif, func_ptr, result, args);

    // If the result is a simple data type smaller than `ffi_arg`, we
    // need to make sure that the type is stored the way the client
    // expects. I think this will make a difference only for big-endian
    // servers.
    // TODO: Does this work for signed numbers?
    if (rtn_type->size < sizeof(ffi_arg))
    {
        switch (rtn_type->size)
        {
            case sizeof(uint8_t):
                *((uint8_t*) result) = *((ffi_arg*) result);
                break;
            case sizeof(uint16_t):
                *((uint16_t*) result) = *((ffi_arg*) result);
                break;
            case sizeof(uint32_t):
                *((uint32_t*) result) = *((ffi_arg*) result);
                break;
            default:
                fprintf(stderr, "[-] Cannot handle size %zd\n",
                        rtn_type->size);
                goto end;
        }
    }

    result = HTONP(result);
    if (send_n(pype->fout, &result, sizeof(result)))
    {
        fputs("[-] Failed to send back result\n", stderr);
        goto end;
    }

    rtn = 0;
end:
    if (rtn && result)
    {
        free(result);
    }
    return rtn;
}

//////////////////////////////////////
// Private function implementations //
//////////////////////////////////////

static ffi_type* recv_type(int fd)
{
    ffi_type* rtn = NULL;
    uint8_t byte_rep;
    if (recv_n(fd, &byte_rep, sizeof(byte_rep)))
    {
        fputs("[-] Failed to receive byte_rep\n", stderr);
        goto end;
    }
    if (byte_rep < ARRAY_LEN(direct_type_conversion_array_g))
    {
        rtn = direct_type_conversion_array_g[byte_rep];
    }
    else if (byte_rep == CP_TYPE_STRUCT)
    {
        fputs("[+] Receiving struct . . .\n", stderr);
        rtn = recv_struct(fd);
        fputs("[+] Done receiving struct!\n", stderr);
    }
end:
    return rtn;
}

static ffi_type* recv_struct(int fd)
{
    ffi_type* rtn = NULL;
    uint32_t num_fields;
    ffi_status status = FFI_BAD_TYPEDEF;

    rtn = calloc(sizeof(*rtn), 1);
    if (!rtn)
    {
        fputs("[-] Failed to alloc struct type\n", stderr);
        goto end;
    }
    rtn->type = FFI_TYPE_STRUCT;

    if (recv_n(fd, &num_fields, sizeof(num_fields)))
    {
        fputs("[-] Failed to receive number of fields\n", stderr);
        goto end;
    }
    num_fields = ntohl(num_fields);
    fprintf(stderr, "    [ ] %d fields\n", num_fields);
    rtn->elements = calloc(num_fields + 1, sizeof(*rtn->elements));
    if (!rtn->elements)
    {
        fputs("[-] Failed to alloc struct fields\n", stderr);
        goto end;
    }
    size_t i;
    for (i = 0; i < num_fields; i++)
    {
        rtn->elements[i] = recv_type(fd);
        if (!rtn->elements[i])
        {
            fprintf(stdout, "[-] Failed to receive field %zd\n", i);
            goto end;
        }
    }
    status = ffi_get_struct_offsets(FFI_DEFAULT_ABI, rtn, NULL);
    if (status != FFI_OK)
    {
        fputs("[-] Failed to layout struct\n", stderr);
        goto end;
    }

end:
    if (status != FFI_OK)
    {
        free_ffi_type(rtn);
        rtn = NULL;
    }
    return rtn;
}

static void free_ffi_type(ffi_type* type)
{
    if (type->type != FFI_TYPE_STRUCT)
    {
        return;
    }
    size_t i;
    // One of the possible failure cases for creating a struct is where
    // we try to alloc `elements` but get back NULL. So we handle that
    // case neatly here.
    if (type->elements)
    {
        for (i = 0; type->elements[i]; i++)
        {
            free_ffi_type(type->elements[i]);
        }
        free(type->elements);
    }
    free(type);
}

static int interpret_abi(uint8_t passed_abi, ffi_abi* f_abi)
{
    int rtn = 0;
    if (passed_abi == CP_ABI_DEFAULT)
    {
        *f_abi = FFI_DEFAULT_ABI;
    }
    else
    {
        rtn = -1;
    }
    return rtn;
}
